from datetime import datetime
from pinthesky.handler import Handler
import logging
import time
import threading


logger = logging.getLogger(__name__)


class CameraThread(threading.Thread, Handler):
    """
    A thread that manages a picamera.PiCamera instance.
    """
    def __init__(
            self, events, sensitivity=10, resolution=(640, 480),
            framerate=20, rotation=270, buffer=15, recording_window=None,
            encoding_bitrate=17000000,
            encoding_profile='high',
            encoding_level="4",
            camera_class=None,
            stream_class=None,
            motion_detection_class=None,
            capture_dir=None):
        super().__init__(daemon=True)
        self.__camera_class = camera_class
        self.__stream_class = stream_class
        self.__motion_detection_class = motion_detection_class
        self.running = True
        self.flushing_stream = False
        self.flushing_ts = None
        self.events = events
        self.buffer = buffer
        self.sensitivity = sensitivity
        self.encoding_bitrate = encoding_bitrate
        self.encoding_profile = encoding_profile
        self.encoding_level = encoding_level
        self.camera = self.__new_camera()
        self.camera.resolution = resolution
        self.camera.framerate = framerate
        self.camera.rotation = rotation
        self.historical_stream = self.__new_stream_buffer()
        self.recording_window = recording_window
        self.capture_dir= capture_dir
        self.configuration_lock = threading.Lock()
        self.__set_recording_window()

    def __set_recording_window(self):
        if self.recording_window is not None:
            self.start_window, self.end_window = map(
                int, self.recording_window.split('-'))

    def __new_motion_detect(self):
        if self.__motion_detection_class is None:
            from pinthesky.motion_detect import MotionDetector
            self.__motion_detection_class = MotionDetector
        return self.__motion_detection_class(
            self.camera, self.events, self.sensitivity)

    def __new_stream_buffer(self):
        if self.__stream_class is None:
            from picamera import PiCameraCircularIO
            self.__stream_class = PiCameraCircularIO
        return self.__stream_class(
            self.camera,
            bitrate=self.encoding_bitrate,
            seconds=self.buffer // 2)

    def __new_camera(self):
        if self.__camera_class is None:
            from picamera import PiCamera
            self.__camera_class = PiCamera
        return self.__camera_class()

    def on_capture_image(self, event):
        logger.info(f'Starting a capture to {self.capture_dir}')
        result = f'{self.capture_dir}/img-{event["timestamp"]}.jpg'
        if 'file_name' in event:
            result = f'{self.capture_dir}/{event["file_name"]}'
        self.camera.capture(result, use_video_port=True)
        self.events.fire_event('capture_image_end', {
            'image_file': result,
            'start_time': event['timestamp']
        })

    def on_motion_start(self, event):
        if not self.flushing_stream:
            logger.info(
                f'Starting a flush on motion event from {event["timestamp"]}')
            self.flushing_ts = event['timestamp']
            self.flushing_stream = True

    def on_file_change(self, event):
        self_fields = [
            "buffer",
            "sensitivity",
            "recording_window",
            "encoding_bitrate",
            "encoding_profile",
            "encoding_level"
        ]
        if "current" in event["content"]:
            cam_obj = event["content"]["current"]["state"]["desired"]["camera"]
            logger.info(f'Update camera fields in {cam_obj}')
            # Hold potentially dangerous mutations if the camera is flushing
            with self.configuration_lock:
                previsouly_recording = self.pause()
                # Update wrapper fields
                for field in self_fields:
                    if field in cam_obj:
                        setattr(self, field, cam_obj[field])
                        if field == "recording_window":
                            self.__set_recording_window()
                # Update picamera fields
                for field in ["rotation", "resolution", "framerate"]:
                    if field in cam_obj:
                        val = cam_obj[field]
                        if field == "resolution":
                            val = tuple(map(int, val.split("x")))
                        setattr(self.camera, field, val)
                if previsouly_recording:
                    self.resume()

    def __flush_video(self):
        # Want to flush when it is safe to flush
        with self.configuration_lock:
            self.camera.split_recording(f'{self.flushing_ts}.after.h264')
            self.historical_stream.copy_to(f'{self.flushing_ts}.before.h264')
            self.historical_stream.clear()
            time.sleep(self.buffer)
            self.camera.split_recording(self.historical_stream)
            self.events.fire_event('flush_end', {
                'start_time': self.flushing_ts
            })
            self.flushing_stream = False

    def run(self):
        logger.info('Starting camera thread')
        self.resume()
        while self.running:
            if not self.flushing_stream and self.recording_window:
                now = datetime.now()
                if now.hour < self.start_window or now.hour > self.end_window:
                    self.pause()
                    time.sleep(1)
                    continue
                else:
                    self.resume()
            self.camera.wait_recording(1)
            if self.flushing_stream:
                self.__flush_video()

    def pause(self):
        if self.camera.recording:
            self.camera.stop_recording()
            logger.info("Camera recording is now paused")
            return True
        return False

    def resume(self):
        if not self.camera.recording:
            self.historical_stream = self.__new_stream_buffer()
            self.camera.start_recording(
                self.historical_stream,
                format='h264',
                bitrate=self.encoding_bitrate,
                profile=self.encoding_profile,
                level=self.encoding_level,
                motion_output=self.__new_motion_detect())
            logger.info("Camera is now recording")
            return True
        return False

    def stop(self):
        self.running = False
        self.camera.stop_recording()
        self.camera.close()



from pinthesky.handler import Handler
import logging
import os

logger = logging.getLogger(__name__)


class VideoCombiner(Handler):
    """
    Combines motion video by concatenating video buffering in memory with
    real-time video. The result of this handle will fire a `combine_end`
    event to signal waiters to do something with the video.
    """
    def __init__(self, events, combine_dir):
        self.events = events
        self.combine_dir = combine_dir

    def on_flush_end(self, event):
        """
        Responds to the camera thread that flushes the videos from buffers
        onto disk. This handle will combine both video parts into a full
        video stream.
        """
        if not os.path.exists(self.combine_dir):
            os.mkdir(self.combine_dir)
        file_name = os.path.join(
            self.combine_dir,
            f'{event["start_time"]}.motion.h264')
        with open(file_name, 'wb') as o:
            for n in ['before', 'after']:
                part_name = f'{event["start_time"]}.{n}.h264'
                with open(part_name, 'rb') as i:
                    o.write(i.read())
                os.remove(part_name)
        self.events.fire_event('combine_end', {
            'start_time': event['start_time'],
            'combine_video': file_name
        })
        logger.info(f'Finish concatinating to {file_name}')



import json

from pinthesky.handler import Handler


class Output(Handler):
    """
    A handler that writes event content to a specified location.
    Watchers on the other end are expected to handle the inotify events.
    """
    def __init__(self, output_file):
        self.output_file = output_file

    def __on_event(self, event):
        with open(self.output_file, 'w') as f:
            f.write(json.dumps(event))

    def on_motion_start(self, event):
        self.__on_event(event)

    def on_combine_end(self, event):
        self.__on_event(event)

    def on_upload_end(self, event):
        self.__on_event(event)

    def on_capture_image_end(self, event):
        self.__on_event(event)



import boto3
import os
import logging

from pinthesky.handler import Handler

logger = logging.getLogger(__name__)


class S3Upload(Handler):
    """
    Handles the `combine_end` to flush the video content to a specific path by
    thing name. Note: if the session is not connected to a remote IoT Thing,
    then this handle does nothing.
    """
    def __init__(
            self, events,
            bucket_name,
            bucket_prefix,
            session,
            bucket_image_prefix=None):
        self.events = events
        self.bucket_name = bucket_name
        self.bucket_prefix = bucket_prefix
        self.session = session
        self.bucket_image_prefix = bucket_image_prefix

    def __upload_to_bucket(self, prefix, file_obj, start_time):
        creds = self.session.login()
        if self.bucket_name is not None and creds is not None:
            video = os.path.basename(file_obj)
            loc = f'{prefix}/{self.session.thing_name}/{video}'
            logger.debug(f"Uploading to s3://{self.bucket_name}/{loc}")
            session = boto3.Session(
                creds['accessKeyId'],
                creds['secretAccessKey'],
                creds['sessionToken'])
            try:
                s3 = session.client('s3')
                with open(file_obj, 'rb') as f:
                    s3.upload_fileobj(f, self.bucket_name, loc)
                self.events.fire_event('upload_end', {
                    'start_time': start_time,
                    'upload': {
                        'bucket_name': self.bucket_name,
                        'bucket_key': loc
                    }
                })
            except RuntimeError as e:
                logger.error(
                    f'Failed to upload to s3://{self.bucket_name}/{loc}: {e}')
            finally:
                # TODO: add a failure strategy / retry attempt here
                os.remove(file_obj)

    def on_capture_image_end(self, event):
        if self.bucket_image_prefix is not None:
            self.__upload_to_bucket(
                self.bucket_image_prefix,
                event['image_file'],
                event['start_time'])

    def on_combine_end(self, event):
        self.__upload_to_bucket(
            self.bucket_prefix,
            event['combine_video'],
            event['start_time'])



from pinthesky.handler import Handler
from requests import get, exceptions
import datetime
import logging
import threading


logger = logging.getLogger(__name__)
fields = [
    "cert_path",
    "key_path",
    "cacert_path",
    "thing_name",
    "role_alias",
    "credentials_endpoint"
]


class Session(Handler):
    """
    An auth session wrapper that caches AWS credential material until expiry.
    """
    def __init__(
            self, cert_path, key_path, cacert_path,
            thing_name, role_alias, credentials_endpoint):
        self.cert_path = cert_path
        self.key_path = key_path
        self.cacert_path = cacert_path
        self.thing_name = thing_name
        self.role_alias = role_alias
        self.credentials = None
        self.refresh_lock = threading.Lock()
        self.__set_endpoint(credentials_endpoint)

    def __set_endpoint(self, endpoint):
        if endpoint is not None and "https://" not in endpoint:
            endpoint = f'https://{endpoint}'
        self.credentials_endpoint = endpoint

    def __is_expired(self, current_time):
        expiration = self.credentials['expiration']
        expiry = datetime.datetime.strptime(expiration, "%Y-%m-%dT%H:%M:%SZ")
        return expiry < current_time

    def on_file_change(self, event):
        if "current" in event["content"]:
            desired = event["content"]["current"]["state"]["desired"]
            con = desired["cloud_connection"]
            with self.refresh_lock:
                for field in fields:
                    if field in con:
                        val = con[field]
                        if field == "credentials_endpoint":
                            self.__set_endpoint(val)
                        else:
                            setattr(self, field, con[field])

    def login(self, force=False):
        ct = datetime.datetime.now()
        if force or self.credentials is None or self.__is_expired(ct):
            with self.refresh_lock:
                try:
                    self.credentials = None
                    if self.credentials_endpoint is not None:
                        res = get(
                            '/'.join([
                                self.credentials_endpoint,
                                'role-aliases',
                                self.role_alias,
                                'credentials']),
                            headers={'x-amzn-iot-thingname': self.thing_name},
                            verify=self.cacert_path,
                            cert=(self.cert_path, self.key_path))
                        res.raise_for_status()
                        payload = res.json()
                        if "credentials" not in payload:
                            logger.warning(
                                f'Expected "credentials", received: {payload}')
                        else:
                            self.credentials = payload['credentials']
                except exceptions.Timeout:
                    logger.error(
                        "Request timeout to %s",
                        self.credentials_endpoint)
                except exceptions.HTTPError as err:
                    logger.error(
                        "Failed to refresh AWS credentials from %s: %s",
                        self.credentials_endpoint,
                        err)
                except exceptions.RequestException:
                    logger.error(
                        "Failed to refresh AWS credentials from %s",
                        self.credentials_endpoint)
        return self.credentials



import picamera.array
import numpy as np


class MotionDetector(picamera.array.PiMotionAnalysis):
    """
    Adapted motion detection class from the PiCamera documentation.
    Performs vactor calculation on a sensitivity threshold.
    """
    def __init__(self, camera, events, sensitivity=10, size=None):
        super(MotionDetector, self).__init__(camera, size)
        self.events = events
        self.sensitivity = sensitivity

    def analyse(self, a):
        a = np.sqrt(
            np.square(a['x'].astype(np.float)) +
            np.square(a['y'].astype(np.float))
            ).clip(0, 255).astype(np.uint8)
        if (a > 60).sum() > self.sensitivity:
            self.events.fire_event('motion_start')



#!/bin/env python3

from setuptools import setup, find_packages


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pinthesky",
    version="0.2.1",
    description="Simple Pi In The Sky device Integration",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Philip Cali",
    author_email="philip.cali@gmail.com",
    url="https://github.com/philcali/pits-device",
    license="Apache License 2.0",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'boto3',
        'requests',
        'numpy',
        'inotify-simple',
        'picamera'
    ],
    extras_require={
        'test': ['pytest', 'requests-mock']
    }
)



import json
import logging
import os
import threading

from pinthesky.handler import Handler
from inotify_simple import INotify, flags

logger = logging.getLogger(__name__)


class INotifyThread(threading.Thread):
    """
    Wrapper class for an INotify watcher. The thread polls events externally
    configured, and pumps changes back into the EventThread. Subscribers are
    notified via the `file_change` event.
    """
    def __init__(self, events, inotify=None):
        super().__init__(daemon=True)
        self.running = True
        self.inotify = inotify or INotify()
        self.events = events
        self.handlers = {}

    def __touch_empty(self, file_name):
        if not os.path.exists(file_name):
            with open(file_name, 'w') as f:
                f.write('{}')

    def __watch_file(self, file_name):
        watch_flags = flags.CREATE | flags.MODIFY
        logger.info(f'Watching input for {file_name}')
        return self.inotify.add_watch(file_name, watch_flags)

    def __fire_event(self, event):
        file_name = None
        for name, wd in self.handlers.items():
            if event.wd == wd:
                file_name = name
        if file_name is not None:
            with open(file_name, 'r') as f:
                content = f.read()
                if content == "":
                    logger.debug(f'The {file_name} was zeroed out. Skipping.')
                    return
                js = json.loads(content)
                self.events.fire_event('file_change', {
                    'file_name': file_name,
                    'content': js
                })
            with open(file_name, 'w') as f:
                f.write("")
            logger.info(f'Zeroing out {file_name} for further use')

    def notify_change(self, file_name):
        if file_name not in self.handlers:
            self.__touch_empty(file_name)
            self.handlers[file_name] = self.__watch_file(file_name)

    def run(self):
        while self.running:
            for event in self.inotify.read():
                mask = flags.from_mask(event.mask)
                if flags.MODIFY in mask or flags.CREATE in mask:
                    self.__fire_event(event)

    def stop(self):
        self.running = False
        for watched in self.watched:
            self.inotify.rm_watch(watched)


class InputHandler(Handler):
    """
    An INotify handler that watches to files who content translate into events.
    """
    def __init__(self, events):
        self.events = events

    def on_file_change(self, event):
        if "name" in event['content'] and "context" in event['content']:
            content = event['content']
            self.events.fire_event(content['name'], content['context'])
